/*
 * Entry point, initialization, miscellaneous routines.
 */

#include <stdio.h>
#include <fcntl.h>

#include "less.h"
#include "position.h"

extern FILE *popen();

public int    ispipe, iscompressed, wascompressed;
public char *	first_cmd = NULL;
public char *	every_first_cmd = NULL;
public int	new_file;
public int	is_tty;
public char *	current_file = NULL;
public char *	previous_file = NULL;
public HANDLE	curr_handle;
public POSITION	initial_pos;
public int	any_display;
public int	scroll;
public int	ac;
public char **	av;
public int 	curr_ac;
public int	quitting;

extern int	file;
extern int	quit_at_eof;
extern int	hit_eof;
extern int	cbufs;
extern int	errmsgs;
extern int	screen_trashed;

static FILE *   compress;
static FILE *   oldcompr;
#if LOGFILE
public int	logfile = -1;
public int	force_logfile = 0;
public char *	namelogfile = NULL;
#endif

#if EDITOR
public char *	editor;
public char *	editproto;
#endif

#if TAGS
extern char *	tagfile;
extern char *	tagpattern;
extern int	tagoption;
#endif

char *myname;
extern char version[];

extern char *strchr();


usage()
{
  int x = strlen(myname);

  printf("\nThis is %s.\n", version + 4);
  printf("\nUsage: %s [-[+]aABcCdeEfgimMnNqQruUsw] [-bN] [-hN] [-xN] [-[z]N]", myname);
  printf("\n       %*s [-P[mM=]string] [-[lL]logfile] [-kkeyfile]", x, "");
  printf("\n       %*s [+cmd] [-ttag] [filename]...\n", x, "");
  printf("\n   or: %s -?\n", myname);
}


int isZfile(name)
char *name;
{
  int res, magic, file;

#ifdef MSDOS
  file = open(name, O_RDONLY | O_BINARY);
#else
  file = open(name, O_RDONLY);
#endif
  res = read(file, &magic, 2);
  close(file);

  return (res == 2) && (magic == 0x9D1F);
}

/*
 * Edit a new file.
 * Filename "-" means standard input.
 * No filename means the "current" file, from the command line.
 */
	public void
edit(filename)
	register char *filename;
{
	register int f;
	register char *m;
	static int didpipe;

	if (filename == NULL || *filename == '\0')
	{
		if (curr_ac >= ac)
		{
			error("No current file");
			return;
		}
		filename = av[curr_ac];
	}

	if (strcmp(filename, "-") == 0)
	{
		/*
		 * Use standard input.
		 */
		if (didpipe)
		{
			error("Can view standard input only once");
			return;
		}
		f = 0;
	} else if ((m = bad_file(filename)) != NULL)
	{
		error(m);
		free(m);
		return;
	}
        else if ( isZfile(filename) )
	{
                char cmd[255];

                sprintf(cmd, "compress -dc %s", filename);
		compress = popen(cmd, "r");
                f = fileno(compress);

		if ( compress == NULL || f < 0 )
		{
                        m = errno_message(filename);
                        error(m);
                        free(m);
			return;
                }

		iscompressed = 1;
        }
#ifdef MSDOS
        else if ((f = open(filename, O_RDONLY | O_BINARY)) < 0)
#else
        else if ((f = open(filename, 0)) < 0)
#endif
	{
		m = errno_message(filename);
		error(m);
		free(m);
		return;
	}

	if (isatty(f))
	{
		/*
		 * Not really necessary to call this an error,
		 * but if the control terminal (for commands)
		 * and the input file (for data) are the same,
		 * we get weird results at best.
		 */
                usage();
                error("\nCan't take input from a terminal !");
		if (f > 0)
			close(f);
		return;
	}

#if LOGFILE
	if (f == 0 && namelogfile != NULL && is_tty)
		use_logfile();
#endif

	/*
	 * We are now committed to using the new file.
	 * Close the current input file and set up to use the new one.
	 */
	if (current_file != NULL)
	{
		store_pos(curr_handle, position(TOP));
		lastmark();
	}

	curr_handle = get_handle(filename);
	initial_pos = recall_pos(curr_handle);

	previous_file = current_file;
	current_file = get_filename(curr_handle);

	if (file > 0)
          if (wascompressed)
            pclose(oldcompr);
          else
		close(file);

	new_file = 1;
        ispipe = (f == 0) || iscompressed;
        wascompressed = iscompressed;
        iscompressed = 0;
	/*
	 * {{ Would this make more sense? }}
	 * ispipe = (lseek(f, (offset_t)0, 0) == BAD_LSEEK);
	 */
	if (ispipe)
		didpipe = 1;
	file = f;
        oldcompr = compress;
	ch_init(cbufs, 0);

	if (every_first_cmd != NULL)
		first_cmd = every_first_cmd;

	if (is_tty)
	{
		int no_display = !any_display;
		any_display = 1;
		/*
		 * Indicate there is nothing displayed yet.
		 */
		pos_clear();
		clr_linenum();
		if (no_display && errmsgs > 0)
		{
			/*
			 * We displayed some messages on error output
			 * (file descriptor 2; see error() function).
			 * Before erasing the screen contents,
			 * display the file name and wait for a keystroke.
			 */
			error(filename);
		}
	}
}

/*
 * Edit the next file in the command line list.
 */
	public void
next_file(n)
	int n;
{
	if (curr_ac + n >= ac)
	{
		if (quit_at_eof && hit_eof)
			quit();
		error("No (N-th) next file");
	} else
		edit(av[curr_ac += n]);
}

/*
 * Edit the previous file in the command line list.
 */
	public void
prev_file(n)
	int n;
{
	if (curr_ac - n < 0)
		error("No (N-th) previous file");
	else
		edit(av[curr_ac -= n]);
}

/*
 * Copy a file directly to standard output.
 * Used if standard output is not a tty.
 */
	static void
cat_file()
{
	register int c;

	while ((c = ch_forw_get()) != EOI)
		putchr(c);
	flush();
}

#if LOGFILE

/*
 * If the user asked for a log file and our input file
 * is standard input, create the log file.
 * We take care not to blindly overwrite an existing file.
 */
use_logfile()
{
	register int exists;
	register int answer;
	register char *m;

	end_logfile();

	/*
	 * {{ We could use access() here. }}
	 */
	exists = open(namelogfile, 0);
	close(exists);
	exists = (exists >= 0);

	if (exists && !force_logfile)
	{
		static char w[] = "WARNING: log file exists: ";
		m = ecalloc(sizeof(w) + strlen(namelogfile), sizeof(char));
		strcpy(m, w);
		strcat(m, namelogfile);
		error(m);
		free(m);
		answer = 'X';	/* Ask the user what to do */
	} else
		answer = 'O';	/* Create the log file */

loop:
	switch (answer)
	{
	case 'O': case 'o':
		logfile = creat(namelogfile, 0644);
		break;
	case 'A': case 'a':
		logfile = open(namelogfile, 1);
		if (lseek(logfile, (offset_t)0, 2) == BAD_LSEEK)
		{
			close(logfile);
			logfile = -1;
		}
		break;
	case 'D': case 'd':
		answer = 0;	/* Don't print an error message */
		break;
	case 'q':
		quit();
	default:
		putstr("\n  Overwrite, Append, or Don't log? ");
		screen_trashed = 1;
		answer = getchr();
		putstr("\n");
		flush();
		goto loop;
	}

	if (logfile < 0 && answer != 0)
	{
		m = ecalloc(strlen(namelogfile) + 20, sizeof(char));
		sprintf(m, "Cannot write to \"%s\"", namelogfile);
		error(m);
		free(m);
	}
}

#endif

/*
 * Entry point.
 */
main(argc, argv)
	int argc;
	char *argv[];
{
	register int i;
	extern char *getenv();

        myname = argv[0];
	/*
	 * Process command line arguments and LESS environment arguments.
	 * Command line arguments override environment arguments.
	 */
	init_prompt();
	init_option();
	scan_option(getenv("LESS"));

#define	isoptstring(s)	(((s)[0] == '-' || (s)[0] == '+') && (s)[1] != '\0')
	argv++;
	while (--argc > 0 && (isoptstring(argv[0]) || isoptpending()))
		scan_option(*argv++);
#undef isoptstring

	if (isoptpending())
	{
		/*
		 * Last command line option was a flag requiring a
		 * following string, but there was no following string.
		 */
		nopendopt();
		exit(0);
	}

#if USERFILE
	/*
	 * Try to use the lesskey file "$HOME/.less".
	 */
	add_hometable();
#endif
#if EDITOR
	editor = getenv("EDITOR");
	if (editor == NULL || *editor == '\0')
		editor = EDIT_PGM;
	editproto = getenv("EDITPROTO");
	if (editproto == NULL || *editproto == '\0')
		editproto = "%E ?lm+%lm. %f";
#endif

	/*
	 * Set up list of files to be examined.
	 */
	ac = argc;
	av = argv;
	curr_ac = 0;

	/*
	 * Set up terminal, etc.
	 */
	is_tty = isatty(1);
	if (!is_tty)
	{
		/*
		 * Output is not a tty.
		 * Just copy the input file(s) to output.
		 */
		if (ac < 1)
		{
			edit("-");
			cat_file();
		} else
		{
			do
			{
				edit((char *)NULL);
				if (file >= 0)
					cat_file();
			} while (++curr_ac < ac);
		}
		exit(0);
	}

	/*
	 * Call save_handle with all the command line filenames,
	 * just to avoid copying them to calloc'ed space later
	 * when we call get_handle() from edit().
	 */
	for (i = 0;  i < ac;  i++)
		save_handle(av[i]);

	init_mark();
	raw_mode(1);
	get_term();
	open_getchr();
	init();

	init_signals(1);

	/*
	 * Select the first file to examine.
	 */
#if TAGS
	if (tagoption)
	{
		/*
		 * A -t option was given.
		 * Verify that no filenames were also given.
		 * Edit the file selected by the "tags" search,
		 * and search for the proper line in the file.
		 */
		if (ac > 0)
		{
			error("No filenames allowed with -t option");
			quit();
		}
		if (tagfile == NULL)
			quit();
		edit(tagfile);
		if (file < 0)
			quit();
		if (tagsearch())
			quit();
	} else
#endif
	if (ac < 1)
		edit("-");	/* Standard input */
	else
	{
		/*
		 * Try all the files named as command arguments.
		 * We are simply looking for one which can be
		 * opened without error.
		 */
		do
		{
			edit((char *)NULL);
		} while (file < 0 && ++curr_ac < ac);
	}

	if (file >= 0)
		commands();
	quit();
	/*NOTREACHED*/
}

/*
 * Copy a string, truncating to the specified length if necessary.
 * Unlike strncpy(), the resulting string is guaranteed to be null-terminated.
 */
	public void
strtcpy(to, from, len)
	char *to;
	char *from;
	unsigned int len;
{
	strncpy(to, from, len);
	to[len-1] = '\0';
}

/*
 * Copy a string to a "safe" place
 * (that is, to a buffer allocated by calloc).
 */
	public char *
save(s)
	char *s;
{
	register char *p;

	p = (char *) ecalloc(strlen(s)+1, sizeof(char));
	strcpy(p, s);
	return (p);
}

	public char *
ecalloc(count, size)
	int count;
	unsigned int size;
{
	register char *p;

	p = calloc(count, size);
	if (p != NULL)
		return (p);
	error("Cannot allocate memory");
	quit();
	/*NOTREACHED*/
}

/*
 * Skip leading spaces in a string.
 */
	public char *
skipsp(s)
	register char *s;
{
	while (*s == ' ' || *s == '\t')
		s++;
	return (s);
}

/*
 * Exit the program.
 */
	public void
quit()
{
	/*
	 * Put cursor at bottom left corner, clear the line,
	 * reset the terminal modes, and exit.
	 */
	quitting = 1;
#if LOGFILE
	end_logfile();
#endif

	if (any_display)
	{        
	lower_left();
	clear_eol();
	deinit();
	flush();

          if (file > 0)
            if (wascompressed)
              pclose(oldcompr);
            else
              close(file);
 	}

	raw_mode(0);
	exit(0);
}
