/*
 * The option table.
 */

#include "less.h"
#include "option.h"

#define	toupper(c)	((c)-'a'+'A')

/*
 * Variables controlled by command line options.
 */
public int seven_bit;		/* Force seven-bit characters? */
public int quiet;		/* Should we suppress the audible bell? */
public int how_search;		/* Where should forward searches start? */
public int top_scroll;		/* Repaint screen from top?
				   (alternative is scroll from bottom) */
public int pr_type;		/* Type of prompt (short, medium, long) */
public int bs_mode;		/* How to process backspaces */
public int know_dumb;		/* Don't complain about dumb terminals */
public int quit_at_eof;		/* Quit after hitting end of file twice */
public int squeeze;		/* Squeeze multiple blank lines into one */
public int tabstop;		/* Tab settings */
public int back_scroll;		/* Repaint screen on backwards movement */
public int twiddle;		/* Display "~" for lines after EOF */
public int caseless;		/* Do "caseless" searches */
public int linenums;		/* Use line numbers */
public int cbufs;		/* Current number of buffers */
public int autobuf;		/* Automatically allocate buffers as needed */
public int nohelp;		/* Disable the HELP command */
public int sendctl;		/* Send control chars to screen untranslated */
public int force_open;		/* Open the file even if not regular file */
public int swindow;

/*
 * Table of all options and their semantics.
 */
static struct option option[] =
{
	{ 'a', TRIPLE, 0, &how_search, NULL,
		"Forward search starts at second REAL line displayed",
		"Forward search starts at bottom of screen",
		"Forward search starts at second SCREEN line displayed"
	},
	{ 'b', NUMBER, 10, &cbufs, opt_b,
		"%d buffers",
		NULL, NULL
	},
	{ 'B', BOOL, 1, &autobuf, NULL,
		"Don't automatically allocate buffers",
		"Automatically allocate buffers when needed",
		NULL
	},
	{ 'c', TRIPLE, 0, &top_scroll, NULL,
		"Repaint by scrolling from bottom of screen",
		"Repaint by clearing each line",
		"Repaint by painting from top of screen"
	},
	{ 'd', BOOL|NO_TOGGLE, 0, &know_dumb, NULL,
		"Assume intelligent terminal",
		"Assume dumb terminal",
		NULL
	},
	{ 'e', TRIPLE, 0, &quit_at_eof, NULL,
		"Don't quit at end-of-file",
		"Quit at end-of-file",
		"Quit immediately at end-of-file"
	},
	{ 'f', BOOL, 0, &force_open, NULL,
		"Open only regular files",
		"Open even non-regular files",
		NULL
	},
	{ 'g', BOOL, 0, &seven_bit, opt_g,
		"Use eight bit characters",
		"Use seven bit characters",
		NULL
	},
	{ 'h', NUMBER, -1, &back_scroll, NULL,
		"Backwards scroll limit is %d lines",
		NULL, NULL
	},
	{ 'H', BOOL|NO_TOGGLE, 0, &nohelp, NULL,
		"Allow help command",
		"Don't allow help command",
		NULL
	},
	{ 'i', BOOL, 0, &caseless, NULL,
		"Case is significant in searches",
		"Ignore case in searches",
		NULL
	},
#if USERFILE
	{ 'k', STRING|NO_TOGGLE, 0, NULL, opt_k,
		NULL, NULL, NULL
	},
#endif
#if LOGFILE
	{ 'l', STRING, 0, NULL, opt_l,
		"log file: ", NULL, NULL
	},
	{ 'L', STRING, 0, NULL, opt__L,
		"Log file: ", NULL, NULL
	},
#endif
	{ 'm', TRIPLE, 0, &pr_type, NULL,
		"Short prompt",
		"Medium prompt",
		"Long prompt"
	},
	{ 'n', TRIPLE|REPAINT, 1, &linenums, NULL,
		"Don't use line numbers",
		"Use line numbers",
		"Constantly display line numbers"
	},
	{ 'P', STRING, 0, NULL, opt_P,
		"prompt: ", NULL, NULL
	},
	{ 'q', TRIPLE, 0, &quiet, NULL,
		"Ring the bell for errors AND at eof/bof",
		"Ring the bell for errors but not at eof/bof",
		"Never ring the bell"
	},
	{ 'r', BOOL|REPAINT, 0, &sendctl, NULL,
		"Control characters are translated",
		"Control characters displayed directly",
		NULL
	},
	{ 's', BOOL|REPAINT, 0, &squeeze, NULL,
		"Don't squeeze multiple blank lines",
		"Squeeze multiple blank lines",
		NULL
	},
#if TAGS
	{ 't', STRING, 0, NULL, opt_t,
		"tag: ", NULL, NULL
	},
#endif
	{ 'u', TRIPLE|REPAINT, 0, &bs_mode, NULL,
		"Underlined text displayed in underline mode",
		"Backspaces cause overstrike",
		"Backspaces print as ^H"
	},
	{ 'w', BOOL|REPAINT, 1, &twiddle, NULL,
		"Display nothing for lines after end-of-file",
		"Display ~ for lines after end-of-file",
		NULL
	},
	{ 'x', NUMBER|REPAINT, 8, &tabstop, NULL,
		"Tab stops every %d spaces",
		NULL, NULL

	},
	{ 'z', NUMBER, -1, &swindow, NULL,
		"Scroll window size is %d lines",
		NULL, NULL
	},
	{ '?', NOVAR, 0, NULL, opt_query,
		NULL, NULL, NULL
	},
	{ '\0' }
};


/*
 * Initialize each option to its default value.
 */
	public void
init_option()
{
	register struct option *o;

	for (o = option;  o->oletter != '\0';  o++)
	{
		/*
		 * Set each variable to its default.
		 */
		if (o->ovar != NULL)
			*(o->ovar) = o->odefault;
	}
}

/*
 * Find an option in the option table.
 */
	public struct option *
findopt(c)
	int c;
{
	register struct option *o;

	for (o = option;  o->oletter != '\0';  o++)
	{
		if (o->oletter == c)
			return (o);
		if ((o->otype & TRIPLE) && toupper(o->oletter) == c)
			return (o);
	}
	return (NULL);
}
