/*
From tumuc!ztivax!unido!mcsun!uunet!allbery Wed Jan 24 14:41:08 1990
Article 364 of comp.sources.misc:
Path: lan!tumuc!ztivax!unido!mcsun!uunet!allbery
From: towfiq@interlan.interlan.com (Mark Mehdi Towfigh)
Newsgroups: comp.sources.misc
Subject: v10i026: Anonymous Pipes in OS/2
Summary: emulation of the bsd unix pipe(2), popen(3), and pclose(3) calls
Keywords: pipe, handles, fork, execl, spawn, popen, pclose
Message-ID: <77161@uunet.UU.NET>
Date: 20 Jan 90 01:17:38 GMT
Sender: allbery@uunet.UU.NET
Organization: Racal InterLan Inc., Boxborough, MA (1-800-LAN-TALK)
Lines: 167
Approved: allbery@uunet.UU.NET (Brandon S. Allbery - comp.sources.misc)
Posted: Sat Jan 20 02:17:38 1990

Posting-number: Volume 10, Issue 26
Submitted-by: towfiq@interlan.interlan.com (Mark Mehdi Towfigh)
Archive-name: pipes.os2

Several people in the past have asked about having Unix-like pipe
calls in OS/2.  The following source file, adapted from 4.3 BSD Unix,
uses a #define to give you a pipe(2) call, and contains function
definitions for popen(3) and pclose(3).  Anyone with problems should
send mail to me; they seem to work fine.

Mark Towfigh
Racal Interlan, Inc.

Modified for bound programs by
Kai Uwe Rommel
(requires DOSMAKEPIPE and DOSCWAIT to be bound to BadDynLink)

----------------------------------cut-here------------------------------------
*/

/*
 * The following code segment is derived from BSD 4.3 Unix.  See
 * copyright below.  Any bugs, questions, improvements, or problems
 * should be sent to Mark Towfigh (towfiq@interlan.interlan.com).
 *
 * Racal InterLan Inc.
 *
 * Description of the three functions contained in this file, adapted
 * from the BSD 4.3 Manual Pages:
 *
 * SYNOPSIS
 *      pipe(fildes)
 *      int fildes[2];
 *
 * DESCRIPTION
 *      The pipe call creates an I/O mechanism called a pipe.
 *      The file descriptors returned can be used in read and write
 *      operations.  When the pipe is written using the descriptor
 *      fildes[1] up to 4096 bytes of data are buffered before the
 *      writing process is suspended.  A read using the descriptor
 *      fildes[0] will pick up the data.
 *
 *      It is assumed that after the pipe has been set up, two (or
 *      more) cooperating processes (created by subsequent spawn
 *      calls) will pass data through the pipe with read and write
 *      calls.
 *
 *      Read calls on an empty pipe (no buffered data) with only one
 *      end (all write file descriptors closed) returns EOF.
 *
 * RETURN VALUE
 *      The function value zero is returned if the pipe was created;
 *      -1 if an error occurred.
 *
 * ERRORS
 *      The pipe call will fail if:
 *
 *      [EMFILE]       Too many descriptors are active.
 *
 *      [ENFILE]       The system file table is full.
 *
 *      [EFAULT]       The fildes buffer is in an invalid area of
 *                     the process's address space.
 *
 * SYNOPSIS
 *
 *      FILE *popen(command, type)
 *      char *command, *type;
 *
 *      pclose(stream)
 *      FILE *stream;
 *
 * DESCRIPTION
 *      The arguments to popen are pointers to null-terminated
 *      strings containing respectively a command line and an
 *      I/O mode, either "r" for reading or "w" for writing.  It
 *      creates a pipe between the calling process and the command
 *      to be executed.  The value returned is a stream pointer that
 *      can be used (as appropriate) to write to the standard input
 *      of the command or read from its standard output.
 *
 *      A stream opened by popen should be closed by pclose, which
 *      waits for the associated process to terminate and returns
 *      the exit status of the command.
 *
 *      Because open files are shared, a type "r" command may be
 *      used as an input filter, and a type "w" as an output filter.
 *
 * DIAGNOSTICS
 *      Popen returns a null pointer if files or processes cannot be
 *      created.
 *
 *      Pclose returns the exit status of the command.
 */

/*
 * Copyright (c) 1980 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <string.h>
#include <process.h>
#include <errno.h>

#define INCL_NOPM
#define	INCL_DOS
#include <os2.h>

static FILE *_popen(char *cmd, char *flags);
static int _pclose(FILE *pipe);

/*
 * emulate Unix pipe(2) call
 */

#define	pipe(x)		DosMakePipe((PHFILE) &x[0], (PHFILE) &x[1], 4096)

#define	tst(a,b)	(*mode == 'r'? (b) : (a))
#define READH           0
#define WRITEH          1

static  int       popen_pid[20];

FILE *popen(char *cmd, char *mode)
{
	int p[2];
        register myside, hisside, save_stream;
        char *arg;

        if ( _osmode == DOS_MODE )
          return _popen(cmd, mode);

	if (pipe(p) < 0)
                return NULL;

        myside = tst(p[WRITEH], p[READH]);
        hisside = tst(p[READH], p[WRITEH]);

	/* set up file descriptors for remote function */
	save_stream = dup(tst(0, 1));		/* don't lose stdin/out! */
        if (dup2(hisside, tst(0, 1)) < 0)
        {
		perror("dup2");
		return NULL;
	}
        close(hisside);

	/*
	 * make sure that we can close our side of the pipe, by
	 * preventing it from being inherited!
	 */

	/* set no-inheritance flag */
	DosSetFHandState(myside, OPEN_FLAGS_NOINHERIT);

        /* separate command name and arguments */
        if ((arg = strchr(cmd, ' ')) != NULL)
          *arg++ = 0;

	/* execute the command:  it will inherit our other file descriptors */
        popen_pid[myside] = spawnlp(P_NOWAIT, cmd, cmd, arg, NULL);

	/* now restore our previous file descriptors */
        if (dup2(save_stream, tst(0, 1)) < 0)   /* retrieve stdin/out */
        {
		perror("dup2");
		return NULL;
	}
        close(save_stream);

	return fdopen(myside, mode);		/* return a FILE pointer */
}

int pclose(FILE *ptr)
{
	register f;
        int status;

        if ( _osmode == DOS_MODE )
          return _pclose(ptr);

	f = fileno(ptr);
        fclose(ptr);

	/* wait for process to terminate */
	cwait(&status, popen_pid[f], WAIT_GRANDCHILD);

	return status;
}


/* a simulation for the Unix popen() and pclose() calls on MS-DOS */
/* only one pipe can be open at a time */

static char pipename[128], command[128];
static int wrpipe;

extern void Mktemp(char *);

static FILE *_popen(char *cmd, char *flags)
{
  wrpipe = (strchr(flags, 'w') != NULL);

  if ( wrpipe )
  {
    strcpy(command, cmd);
    strcpy(pipename, "~WXXXXXX");
    Mktemp(pipename);
    return fopen(pipename, flags);  /* ordinary file */
  }
  else
  {
    strcpy(pipename, "~RXXXXXX");
    Mktemp(pipename);
    strcpy(command, cmd);
    strcat(command, ">");
    strcat(command, pipename);
    system(command);
    return fopen(pipename, flags);  /* ordinary file */
  }
}

static int _pclose(FILE *pipe)
{
  int rc;

  if ( fclose(pipe) == EOF )
    return EOF;

  if ( wrpipe )
  {
    if ( command[strlen(command) - 1] == '!' )
      command[strlen(command) - 1] = 0;
    else
      strcat(command, "<");

    strcat(command, pipename);
    rc = system(command);
    unlink(pipename);
    return rc;
  }
  else
  {
    unlink(pipename);
    return 0;
  }
}
